# -*- coding: utf-8 -*-

# ******************************************************************************
#
# Copyright (C) 2006-2011 Olivier Tilloy <olivier@tilloy.net>
# Copyright (C) 2015-2023 Vincent Vande Vyvre <vincent.vandevyvre@oqapy.eu>
#
# This file is part of the py3exiv2 distribution.
#
# py3exiv2 is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 3 as published by the Free Software Foundation.
#
# py3exiv2 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with py3exiv2; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, 5th Floor, Boston, MA 02110-1301 USA.
#
# Maintainer: Vincent Vande Vyvre <vincent.vandevyvre@oqapy.eu>
#
# ******************************************************************************

"""
IPTC specific code.
"""

import libexiv2python

from pyexiv2.utils import ListenerInterface, NotifyingList, \
                          FixedOffset, DateTimeFormatter

import time
import datetime
import re
import warnings


class IptcValueError(ValueError):

    """
    Exception raised when failing to parse the *value* of an IPTC tag.

    :attribute value: the value that fails to be parsed
    :type value: string
    :attribute type: the IPTC type of the tag
    :type type: string
    """

    def __init__(self, value, type):
        self.value = value
        self.type = type

    def __str__(self):
        return 'Invalid value for IPTC type [%s]: [%s]' % \
               (self.type, self.value)


class IptcTag(ListenerInterface):
    """An IPTC tag.

    This tag can have several values (tags that have the *repeatable* property).

    Here is a correspondance table between the IPTC types and the possible
    python types the value of a tag may take:

    - Short: int
    - String: string
    - Date: :class:`datetime.date`
    - Time: :class:`datetime.time`
    - Undefined: string
    """
    # strptime is not flexible enough to handle all valid Time formats, we use a
    # custom regular expression
    _time_zone_re = r'(?P<sign>\+|-)(?P<ohours>\d{2}):(?P<ominutes>\d{2})'
    _time_re = re.compile(r'(?P<hours>\d{2}):(?P<minutes>\d{2}):(?P<seconds>\d{2})(?P<tzd>%s)' % _time_zone_re)
    def __init__(self, key, values=None, _tag=None):
        """The tag can be initialized with an optional list of values which
        expected type depends on the IPTC type of the tag.

        Args:
        key -- the key of the tag
        values -- the values of the tag
        """
        super(IptcTag, self).__init__()
        if _tag is not None:
            self._tag = _tag

        else:
            self._tag = libexiv2python._IptcTag(key)

        self._raw_values = None
        self._values = None
        self._values_cookie = False
        if values is not None:
            self._set_values(values)

    def _set_owner(self, metadata):
        self._tag._setParentImage(metadata._image)

    @staticmethod
    def _from_existing_tag(_tag):
        # Build a tag from an already existing libexiv2python._IptcTag
        tag = IptcTag(_tag._getKey(), _tag=_tag)
        # Do not set the raw_value property, as it would call
        # _tag._setRawValues
        # (see https://bugs.launchpad.net/pyexiv2/+bug/582445).
        tag._raw_values = _tag._getRawValues()
        tag._values_cookie = True
        return tag

    @property
    def key(self):
        """The key of the tag in the dotted form
        ``familyName.groupName.tagName`` where ``familyName`` = ``iptc``.

        """
        return self._tag._getKey()

    @property
    def type(self):
        """The IPTC type of the tag (one of Short, String, Date, Time,
        Undefined).

        """
        return self._tag._getType()

    @property
    def name(self):
        """The name of the tag (this is also the third part of the key).

        """
        return self._tag._getName()

    @property
    def title(self):
        """The title (label) of the tag.

        """
        return self._tag._getTitle()

    @property
    def description(self):
        """The description of the tag.

        """
        return self._tag._getDescription()

    @property
    def photoshop_name(self):
        """The Photoshop name of the tag.

        """
        return self._tag._getPhotoshopName()

    @property
    def repeatable(self):
        """Whether the tag is repeatable (accepts several values).

        """
        return self._tag._isRepeatable()

    @property
    def record_name(self):
        """The name of the tag's record.

        """
        return self._tag._getRecordName()

    @property
    def record_description(self):
        """The description of the tag's record.

        """
        return self._tag._getRecordDescription()

    def _get_raw_values(self):
        return self._raw_values

    def _set_raw_values(self, values):
        if not isinstance(values, (list, tuple)):
            raise TypeError('Expecting a list of values')

        self._tag._setRawValues(values)
        self._raw_values = values
        self._values_cookie = True

    raw_value = property(fget=_get_raw_values, fset=_set_raw_values,
                         doc='The raw values of the tag as a list of strings.')

    def _compute_values(self):
        # Lazy computation of the values from the raw values
        self._values = NotifyingList([self._convert_to_python(v) 
                                        for v in self._raw_values])
        self._values.register_listener(self)
        self._values_cookie = False

    def _get_values(self):
        if self._values_cookie:
            self._compute_values()

        return self._values

    def _set_values(self, values):
        if not isinstance(values, (list, tuple)):
            raise TypeError('Expecting a list of values')

        self.raw_value = [self._convert_to_string(v) for v in values]

        if isinstance(self._values, NotifyingList):
            self._values.unregister_listener(self)

        if isinstance(values, NotifyingList):
            # Already a notifying list
            self._values = values

        else:
            # Make the values a notifying list 
            self._values = NotifyingList(values)

        self._values.register_listener(self)
        self._values_cookie = False

    value = property(fget=_get_values, fset=_set_values,
                     doc='The values of the tag as a list of python objects.')

    def contents_changed(self):
        # Implementation of the ListenerInterface.
        # React on changes to the list of values of the tag.
        # The contents of self._values was changed.
        # The following is a quick, non optimal solution.
        self._set_values(self._values)

    def _convert_to_python(self, value):
        """Convert one raw value to its corresponding python type.

        Args:
        value -- the raw value to be converted

        Return: the value converted to its corresponding python type

        Raise IptcValueError: if the conversion fails
        """
        if self.type == 'Short':
            try:
                return int(value)
            except ValueError:
                raise IptcValueError(value, self.type)

        elif self.type == 'String':
            # There is currently no charset conversion.
            # TODO: guess the encoding and decode accordingly into unicode
            # where relevant.
            if isinstance(value, bytes):
                try:
                    value = value.decode('utf-8')
                except UnicodeDecodeError:
                    # Unknow encoding, return the raw value
                    pass
            return value

        elif self.type == 'Date':
            # According to the IPTC specification, the format for a string field
            # representing a date is '%Y%m%d'. However, the string returned by
            # exiv2 using method DateValue::toString() is formatted using
            # pattern '%Y-%m-%d'.
            format = '%Y-%m-%d'
            try:
                t = time.strptime(value, format)
                return datetime.date(*t[:3])
            except ValueError:
                raise IptcValueError(value, self.type)

        elif self.type == 'Time':
            # According to the IPTC specification, the format for a string field
            # representing a time is '%H%M%S±%H%M'. However, the string returned
            # by exiv2 using method TimeValue::toString() is formatted using
            # pattern '%H:%M:%S±%H:%M'.
            match = IptcTag._time_re.match(value)
            if match is None:
                raise IptcValueError(value, self.type)

            gd = match.groupdict()
            try:
                tzinfo = FixedOffset(gd['sign'], int(gd['ohours']),
                                     int(gd['ominutes']))
            except TypeError:
                raise IptcValueError(value, self.type)

            try:
                return datetime.time(int(gd['hours']), int(gd['minutes']),
                                     int(gd['seconds']), tzinfo=tzinfo)
            except (TypeError, ValueError):
                raise IptcValueError(value, self.type)

        elif self.type == 'Undefined':
            # Binary data, return it unmodified
            return value

        raise IptcValueError(value, self.type)

    def _convert_to_string(self, value):
        """Convert one value to its corresponding string representation,
        suitable to pass to libexiv2.

        Args:
        value -- the value to be converted

        Return: the value converted to its corresponding string representation

        Raise IptcValueError: if the conversion fails
        """
        if self.type == 'Short':
            if isinstance(value, int):
                return str(value)

            else:
                raise IptcValueError(value, self.type)

        elif self.type == 'String':
            if isinstance(value, str):
                try:
                    return value.encode('utf-8')
                except UnicodeEncodeError:
                    raise IptcValueError(value, self.type)

            elif isinstance(value, bytes):
                return value

            else:
                raise IptcValueError(value, self.type)

        elif self.type == 'Date':
            if isinstance(value, (datetime.date, datetime.datetime)):
                return DateTimeFormatter.iptc_date(value)

            else:
                raise IptcValueError(value, self.type)

        elif self.type == 'Time':
            if isinstance(value, (datetime.time, datetime.datetime)):
                return DateTimeFormatter.iptc_time(value)

            else:
                raise IptcValueError(value, self.type)

        elif self.type == 'Undefined':
            if isinstance(value, str):
                return value

            else:
                raise IptcValueError(value, self.type)

        raise IptcValueError(value, self.type)

    def __str__(self):
        """Return a string representation of the IPTC tag for debugging purposes

        """
        left = '%s [%s]' % (self.key, self.type)
        if self._raw_values is None:
            right = '(No values)'

        else:
             right = self._raw_values

        return '<%s = %s>' % (left, right)

    # Support for pickling.
    def __getstate__(self):
        return (self.key, self.raw_value)

    def __setstate__(self, state):
        key, raw_value = state
        self._tag = libexiv2python._IptcTag(key)
        self.raw_value = raw_value




# -*- coding: utf-8 -*-

# ******************************************************************************
#
# Copyright (C) 2006-2011 Olivier Tilloy <olivier@tilloy.net>
# Copyright (C) 2015-2023 Vincent Vande Vyvre <vincent.vandevyvre@oqapy.eu>
#
# This file is part of the py3exiv2 distribution.
#
# py3exiv2 is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 3 as published by the Free Software Foundation.
#
# py3exiv2 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with py3exiv2; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, 5th Floor, Boston, MA 02110-1301 USA.
#
# Maintainer: Vincent Vande Vyvre <vincent.vandevyvre@oqapy.eu>
#
# ******************************************************************************

"""
EXIF specific code.
"""

import libexiv2python

from pyexiv2.utils import (is_fraction, make_fraction, fraction_to_string,
                          NotifyingList, ListenerInterface,
                          undefined_to_string, string_to_undefined,
                          DateTimeFormatter)

import time
import datetime
import sys

class ExifValueError(ValueError):
    """Exception raised when failing to parse the *value* of an EXIF tag.

    """
    def __init__(self, value, type_):
        """Instanciate the ExifValueError.

        Args:
        value -- the value that fails to be parsed
        type_ -- the EXIF type of the tag
        """
        self.value = value
        self.type = type_

    def __str__(self):
        return 'Invalid value for EXIF type [%s]: [%s]' %(self.type, self.value)


class ExifTag(ListenerInterface):
    """An EXIF tag.

    Here is a correspondance table between the EXIF types and the possible
    python types the value of a tag may take:

    - Ascii: :class:`datetime.datetime`, :class:`datetime.date`, string
    - Byte, SByte: bytes
    - Comment: string
    - Long, SLong: [list of] int
    - Short, SShort: [list of] int
    - Rational, SRational: [list of] :class:`fractions.Fraction` if available
      (Python ≥ 2.6) or :class:`pyexiv2.utils.Rational`      
    - Undefined: string
    """
    # According to the EXIF specification, the only accepted format for an Ascii
    # value representing a datetime is '%Y:%m:%d %H:%M:%S', but it seems that
    # others formats can be found in the wild.
    _datetime_formats = ('%Y:%m:%d %H:%M:%S',
                         '%Y-%m-%d %H:%M:%S',
                         '%Y-%m-%dT%H:%M:%SZ')

    _date_formats = ('%Y:%m:%d',)

    def __init__(self, key, value=None, _tag=None):
        """ The tag can be initialized with an optional value which expected
        type depends on the EXIF type of the tag.

        Args:
        key -- the key of the tag
        value -- the value of the tag
        """
        super().__init__()
        if _tag is not None:
            self._tag = _tag

        else:
            self._tag = libexiv2python._ExifTag(key)

        self._raw_value = None
        self._value = None
        self._value_cookie = False
        if value is not None:
            self._set_value(value)

    def _set_owner(self, metadata):
        self._tag._setParentImage(metadata._image)

    @staticmethod
    def _from_existing_tag(_tag):
        """Build a tag from an already existing libexiv2python._ExifTag.

        """
        tag = ExifTag(_tag._getKey(), _tag=_tag)
        # Do not set the raw_value property, as it would call _tag._setRawValue
        # (see https://bugs.launchpad.net/pyexiv2/+bug/582445).
        tag._raw_value = _tag._getRawValue()
        tag._value_cookie = True
        return tag

    @property
    def key(self):
        """The key of the tag in the dotted form
        ``familyName.groupName.tagName`` where ``familyName`` = ``exif``.

        """
        return self._tag._getKey()

    @property
    def type(self):
        """The EXIF type of the tag (one of Ascii, Byte, SByte, Comment, Short,
        SShort, Long, SLong, Rational, SRational, Undefined).

        """
        return self._tag._getType()

    @property
    def name(self):
        """The name of the tag (this is also the third part of the key).

        """
        return self._tag._getName()

    @property
    def label(self):
        """The title (label) of the tag.

        """
        return self._tag._getLabel()

    @property
    def description(self):
        """The description of the tag.

        """
        return self._tag._getDescription()

    @property
    def section_name(self):
        """The name of the tag's section.

        """
        return self._tag._getSectionName()

    @property
    def section_description(self):
        """The description of the tag's section.

        """
        return self._tag._getSectionDescription()

    def _get_raw_value(self):
        return self._raw_value

    def _set_raw_value(self, value):
        self._tag._setRawValue(value)
        self._raw_value = value
        self._value_cookie = True

    raw_value = property(fget=_get_raw_value, fset=_set_raw_value,
                         doc='The raw value of the tag as a string.')

    def _compute_value(self):
        """Lazy computation of the value from the raw value.

        """
        if self.type in ('Short', 'SShort', 'Long', 'SLong', 
                         'Rational', 'SRational'):
            # May contain multiple values
            values = self._raw_value.split()
            if len(values) > 1:
                # Make values a notifying list
                values = [self._convert_to_python(v) for v in values]
                self._value = NotifyingList(values)
                self._value.register_listener(self)
                self._value_cookie = False
                return

        self._value = self._convert_to_python(self._raw_value)
        self._value_cookie = False

    def _get_value(self):
        if self._value_cookie:
            self._compute_value()
        return self._value

    def _set_value(self, value):
        if isinstance(value, (list, tuple)):
            raw_values = [self._convert_to_string(v) for v in value]
            self.raw_value = ' '.join(raw_values)

        else:
            self.raw_value = self._convert_to_string(value)

        if isinstance(self._value, NotifyingList):
            self._value.unregister_listener(self)

        if isinstance(value, NotifyingList):
            # Already a notifying list
            self._value = value
            self._value.register_listener(self)

        elif isinstance(value, (list, tuple)):
            # Make the values a notifying list 
            self._value = NotifyingList(value)
            self._value.register_listener(self)

        else:
            # Single value
            self._value = value

        self._value_cookie = False

    value = property(fget=_get_value, fset=_set_value,
                     doc='The value of the tag as a python object.')

    @property
    def human_value(self):
        """A (read-only) human-readable representation
        of the value of the tag.

        """
        return self._tag._getHumanValue() or None

    def contents_changed(self):
        # Implementation of the ListenerInterface.
        # React on changes to the list of values of the tag.
        # self._value is a list of values and its contents changed.
        self._set_value(self._value)

    def _match_encoding(self, charset):
        # charset see:
        # http://www.exiv2.org/doc/classExiv2_1_1CommentValue.html
        # enum  	CharsetId {
        #           ascii, jis, unicode, undefined,
        #           invalidCharsetId, lastCharsetId } 
        encoding = sys.getdefaultencoding()
        if charset in ('Ascii', 'ascii'):
            encoding = 'ascii'

        elif charset in ('Jis', 'jis'):
            encoding = 'shift_jis'

        elif charset in ('Unicode', 'unicode'):
            encoding = 'utf-8'

        return encoding

    def _convert_to_python(self, value):
        """
        Convert one raw value to its corresponding python type.

        :param value: the raw value to be converted
        :type value: string

        :return: the value converted to its corresponding python type

        :raise ExifValueError: if the conversion fails
        """
        if self.type == 'Ascii':
            # The value may contain a Datetime
            for format in self._datetime_formats:
                try:
                    t = time.strptime(value, format)
                except ValueError:
                    continue
                else:
                    return datetime.datetime(*t[:6])
            # Or a Date (e.g. Exif.GPSInfo.GPSDateStamp)
            for format in self._date_formats:
                try:
                    t = time.strptime(value, format)
                except ValueError:
                    continue
                else:
                    return datetime.date(*t[:3])
            # Default to string.
            # There is currently no charset conversion.
            # TODO: guess the encoding and decode accordingly into unicode
            # where relevant.
            return value

        elif self.type in ('Byte', 'SByte'):
            if isinstance(value, bytes):
                return value.decode('utf-8')
            return value

        elif self.type == 'Comment':
            if isinstance(value, str):
                if value.startswith('charset='):
                    charset, val = value.split(' ', 1)
                    return val
                return value

            if value.startswith(b'charset='):          
                charset = charset.split('=')[1].strip('"')
                encoding = self._match_encoding(charset)
                return val.decode(encoding, 'replace')

            else:
                # No encoding defined.
                try:
                    return value.decode('utf-8')
                except UnicodeError:
                    pass

            return value

        elif self.type in ('Short', 'SShort'):
            try:
                return int(value)
            except ValueError:
                raise ExifValueError(value, self.type)

        elif self.type in ('Long', 'SLong'):
            try:
                return int(value)
            except ValueError:
                raise ExifValueError(value, self.type)

        elif self.type in ('Rational', 'SRational'):
            try:
                r = make_fraction(value)
            except (ValueError, ZeroDivisionError):
                raise ExifValueError(value, self.type)

            else:
                if self.type == 'Rational' and r.numerator < 0:
                    raise ExifValueError(value, self.type)
                return r

        elif self.type == 'Undefined':
            # There is currently no charset conversion.
            # TODO: guess the encoding and decode accordingly into unicode
            # where relevant.
            return undefined_to_string(value)

        raise ExifValueError(value, self.type)

    def _convert_to_string(self, value):
        """
        Convert one value to its corresponding string representation, suitable
        to pass to libexiv2.

        :param value: the value to be converted

        :return: the value converted to its corresponding string representation
        :rtype: string

        :raise ExifValueError: if the conversion fails
        """
        if self.type == 'Ascii':
            if isinstance(value, datetime.datetime):
                return DateTimeFormatter.exif(value)

            elif isinstance(value, datetime.date):
                if self.key == 'Exif.GPSInfo.GPSDateStamp':
                    # Special case
                    return DateTimeFormatter.exif(value)

                else:
                    return '%s 00:00:00' % DateTimeFormatter.exif(value)

            else:
                return value

        elif self.type in ('Byte', 'SByte'):
            if isinstance(value, str):
                try:
                    return value.encode('utf-8')
                except UnicodeEncodeError:
                    raise ExifValueError(value, self.type)

            elif isinstance(value, bytes):
                return value

            else:
                raise ExifValueError(value, self.type)

        elif self.type == 'Comment':
            return self._convert_to_bytes(value)

        elif self.type == 'Short':
            if isinstance(value, int) and value >= 0:
                return str(value)

            else:
                raise ExifValueError(value, self.type)

        elif self.type == 'SShort':
            if isinstance(value, int):
                return str(value)

            else:
                raise ExifValueError(value, self.type)

        elif self.type == 'Long':
            if isinstance(value, int) and value >= 0:
                return str(value)

            else:
                raise ExifValueError(value, self.type)

        elif self.type == 'SLong':
            if isinstance(value, int):
                return str(value)

            else:
                raise ExifValueError(value, self.type)

        elif self.type == 'Rational':
            if is_fraction(value) and value.numerator >= 0:
                return fraction_to_string(value)

            else:
                raise ExifValueError(value, self.type)

        elif self.type == 'SRational':
            if is_fraction(value):
                return fraction_to_string(value)

            else:
                raise ExifValueError(value, self.type)

        elif self.type == 'Undefined':
            if isinstance(value, str):
                try:
                    return string_to_undefined(value)
                except UnicodeEncodeError:
                    raise ExifValueError(value, self.type)

            elif isinstance(value, bytes):
                return string_to_undefined(value)

            else:
                raise ExifValueError(value, self.type)

        raise ExifValueError(value, self.type)

    def _convert_to_bytes(self, value):
        if value is None:
            return

        if isinstance(value, str):
            if value.startswith('charset='):
                charset, val = value.split(' ', 1)
                charset = charset.split('=')[1].strip('"')
                encoding = self._match_encoding(charset)

            else:
                encoding = 'utf-8'
                charset = 'Unicode'

            try:
                val = value.encode(encoding)
            except UnicodeError:
                pass

            else:
                #self._set_raw_value('charset=%s %s' % (charset, val))
                return val

        elif isinstance(value, bytes):
            return value

        else:
            raise ExifValueError(value, self.type)

    def __str__(self):
        """
        :return: a string representation of the EXIF tag for debugging purposes
        :rtype: string
        """
        left = '%s [%s]' % (self.key, self.type)
        if self._raw_value is None:
            right = '(No value)'

        elif self.type == 'Undefined' and len(self._raw_value) > 100:
            right = '(Binary value suppressed)'

        else:
             right = self._raw_value

        return '<%s = %s>' % (left, right)

    # Support for pickling.
    def __getstate__(self):
        return (self.key, self.raw_value)

    def __setstate__(self, state):
        key, raw_value = state
        self._tag = libexiv2python._ExifTag(key)
        self.raw_value = raw_value


class ExifThumbnail(object):

    """
    A thumbnail image optionally embedded in the IFD1 segment of the EXIF data.

    The image is either a TIFF or a JPEG image.
    """

    def __init__(self, _metadata):
        self._metadata = _metadata

    @property
    def mime_type(self):
        """The mime type of the preview image (e.g. ``image/jpeg``)."""
        return self._metadata._image._getExifThumbnailMimeType()

    @property
    def extension(self):
        """The file extension of the preview image with a leading dot
        (e.g. ``.jpg``)."""
        return self._metadata._image._getExifThumbnailExtension()

    def write_to_file(self, path):
        """
        Write the thumbnail image to a file on disk.
        The file extension will be automatically appended to the path.

        :param path: path to write the thumbnail to (without an extension)
        :type path: string
        """
        self._metadata._image._writeExifThumbnailToFile(path)

    def _update_exif_tags_cache(self):
        # Update the cache of EXIF tags
        keys = self._metadata._image._exifKeys()
        self._metadata._keys['exif'] = keys
        #cached = self._metadata._tags['exif'].keys()
        for key in self._metadata._tags['exif'].keys():
            if key not in keys:
                del self._metadata._tags['exif'][key]

    def erase(self):
        """
        Delete the thumbnail from the EXIF data.
        Removes all Exif.Thumbnail.*, i.e. Exif IFD1 tags.
        """
        self._metadata._image._eraseExifThumbnail()
        self._update_exif_tags_cache()

    def set_from_file(self, path):
        """
        Set the EXIF thumbnail to the JPEG image path.
        This sets only the ``Compression``, ``JPEGInterchangeFormat`` and
        ``JPEGInterchangeFormatLength`` tags, which is not all the thumbnail
        EXIF information mandatory according to the EXIF standard
        (but it is enough to work with the thumbnail).

        :param path: path to a JPEG file to set the thumbnail to
        :type path: string
        """
        self._metadata._image._setExifThumbnailFromFile(path)
        self._update_exif_tags_cache()

    def _get_data(self):
        buf_ = self._metadata._image._getExifThumbnailData()
        return buf_

    def _set_data(self, data):
        self._metadata._image._setExifThumbnailFromData(data)
        self._update_exif_tags_cache()

    data = property(fget=_get_data, fset=_set_data,
                    doc='The raw thumbnail data. Setting it is restricted to ' +
                        'a buffer in the JPEG format.')




# -*- coding: utf-8 -*-

# ******************************************************************************
#
# Copyright (C) 2006-2011 Olivier Tilloy <olivier@tilloy.net>
# Copyright (C) 2015-2023 Vincent Vande Vyvre <vincent.vandevyvre@oqapy.eu>
#
# This file is part of the py3exiv2 distribution.
#
# py3exiv2 is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 3 as published by the Free Software Foundation.
#
# py3exiv2 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with py3exiv2; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, 5th Floor, Boston, MA 02110-1301 USA.
#
# Maintainer: Vincent Vande Vyvre <vincent.vandevyvre@oqapy.eu>
#
# ******************************************************************************

"""
Provide the ImageMetadata class.
"""

import os
import sys
import codecs

from errno import ENOENT
from itertools import chain

if sys.version_info < (3, 3):
    from collections import MutableMapping
else:
    from collections.abc import MutableMapping

import libexiv2python

from pyexiv2.exif import ExifTag, ExifThumbnail
from pyexiv2.iptc import IptcTag
from pyexiv2.xmp import XmpTag
from pyexiv2.preview import Preview


class ImageMetadata(MutableMapping):
    """A container for all the metadata embedded in an image.

    It provides convenient methods for the manipulation of EXIF, IPTC and XMP
    metadata embedded in image files such as JPEG and TIFF files, using Python
    types.
    It also provides access to the previews embedded in an image.
    """

    def __init__(self, filename):
        """Instanciate the ImageMeatadata class.

        Args:
        filename: str(path to an image file)
        """
        self.filename = filename
        self.__image = None
        self._keys = {'exif': None, 'iptc': None, 'xmp': None}
        self._tags = {'exif': {}, 'iptc': {}, 'xmp': {}}
        self._exif_thumbnail = None

    def _instantiate_image(self, filename):
        """Instanciate the exiv2 image.

        Args:
        filename -- str(path to an image file)
        """
        # This method is meant to be overridden in unit tests to easily replace
        # the internal image reference by a mock.
        if not os.path.exists(filename) or not os.path.isfile(filename):
            raise IOError(ENOENT, os.strerror(ENOENT), filename)

        # Remember the reference timestamps before doing any access to the file
        stat = os.stat(filename)
        self._atime = stat.st_atime
        self._mtime = stat.st_mtime
        return libexiv2python._Image(filename)

    @classmethod
    def from_buffer(cls, buffer_):
        """Instantiate an image container from an image memoryview.

        Args:
        buffer_ -- a memoryview containing image data as bytes
        """
        obj = cls(None)
        obj.__image = libexiv2python._Image(buffer_, len(buffer_))
        return obj

    @property
    def _image(self):
        if self.__image is None:
            raise IOError('Image metadata has not been read yet')

        return self.__image

    def read(self):
        """Read the metadata embedded in the associated image.

        It is necessary to call this method once before attempting to access
        the metadata (an exception will be raised if trying to access metadata
        before calling this method).
        """
        if self.__image is None:
            self.__image = self._instantiate_image(self.filename)

        self.__image._readMetadata()

    def write(self, preserve_timestamps=False):
        """Write the metadata back to the image.

        Args:
        preserve_timestamps -- whether to preserve the file's original
                               timestamps (access time and modification time)
                               Type: boolean
        """
        self._image._writeMetadata()
        if self.filename is None:
            return

        if preserve_timestamps:
            # Revert to the original timestamps
            os.utime(self.filename, (self._atime, self._mtime))

        else:
            # Reset the reference timestamps
            stat = os.stat(self.filename)
            self._atime = stat.st_atime
            self._mtime = stat.st_mtime

    @property
    def dimensions(self):
        """A tuple containing the width and height of the image, expressed in
        pixels.

        """
        return (self._image._getPixelWidth(), self._image._getPixelHeight())

    @property
    def mime_type(self):
        """The mime type of the image, as a string.

        """
        return self._image._getMimeType()

    @property
    def exif_keys(self):
        """Return the list of the keys of the available EXIF tags.

        """
        if self._keys['exif'] is None:
            self._keys['exif'] = self._image._exifKeys()

        return self._keys['exif']

    @property
    def iptc_keys(self):
        """Return the list of the keys of the available IPTC tags.

        """
        if self._keys['iptc'] is None:
            self._keys['iptc'] = self._image._iptcKeys()

        return self._keys['iptc']

    @property
    def xmp_keys(self):
        """Return the list of the keys of the available XMP tags.

        """
        if self._keys['xmp'] is None:
            self._keys['xmp'] = self._image._xmpKeys()

        return self._keys['xmp']

    def _get_exif_tag(self, key):
        """Return the EXIF tag for the given key.

        Throw a KeyError if the tag doesn't exist.

        Args:
        key -- the exif key
        """
        try:
            return self._tags['exif'][key]
        except KeyError:
            _tag = self._image._getExifTag(key)
            tag = ExifTag._from_existing_tag(_tag)
            self._tags['exif'][key] = tag
            return tag

    def _get_iptc_tag(self, key):
        """Return the IPTC tag for the given key.

        Throw a KeyError if the tag doesn't exist.

        Args:
        key -- the iptc key
        """
        try:
            return self._tags['iptc'][key]
        except KeyError:
            _tag = self._image._getIptcTag(key)
            tag = IptcTag._from_existing_tag(_tag)
            self._tags['iptc'][key] = tag
            return tag

    def _get_xmp_tag(self, key):
        """Return the XMP tag for the given key.

        Throw a KeyError if the tag doesn't exist.

        Args:
        key -- the xmp key
        """
        try:
            return self._tags['xmp'][key]
        except KeyError:
            _tag = self._image._getXmpTag(key)
            tag = XmpTag._from_existing_tag(_tag)
            self._tags['xmp'][key] = tag
            return tag

    def __getitem__(self, key):
        """Return a metadata tag for a given key.

        Raise KeyError if the tag doesn't exist

        Args:
        key -- metadata key in the dotted form
               ``familyName.groupName.tagName`` where ``familyName`` may
               be one of ``exif``, ``iptc`` or ``xmp``.
        """
        family = key.split('.')[0].lower()
        if family in ('exif', 'iptc', 'xmp'):
            return getattr(self, '_get_%s_tag' % family)(key)

        else:
            raise KeyError(key)

    def _set_exif_tag(self, key, tag_or_value):
        """Set an EXIF tag. If the tag already exists, its value is overwritten.

        Args:
        key -- the EXIF key
        tag_or_value -- an ExifTag instance or the value of the data
        """
        if isinstance(tag_or_value, ExifTag):
            tag = tag_or_value

        else:
            # As a handy shortcut, accept direct value assignment.
            tag = ExifTag(key, tag_or_value)

        tag._set_owner(self)
        self._tags['exif'][tag.key] = tag
        if tag.key not in self.exif_keys:
            self._keys['exif'].append(tag.key)

    def _set_iptc_tag(self, key, tag_or_values):
        """Set an IPTC tag. If the tag already exists, its value is overwritten.

        Args:
        key -- the IPTC key
        tag_or_value -- an IptcTag instance or the value of the data
        """
        if isinstance(tag_or_values, IptcTag):
            tag = tag_or_values

        else:
            # As a handy shortcut, accept direct value assignment.
            tag = IptcTag(key, tag_or_values)

        tag._set_owner(self)
        self._tags['iptc'][tag.key] = tag
        if tag.key not in self.iptc_keys:
            self._keys['iptc'].append(tag.key)

    def _set_xmp_tag(self, key, tag_or_value):
        """Set an XMP tag. If the tag already exists, its value is overwritten.

        Args:
        key -- the XMP key
        tag_or_value -- an XmpTag instance or the value of the data
        """
        if isinstance(tag_or_value, XmpTag):
            tag = tag_or_value

        else:
            # As a handy shortcut, accept direct value assignment.
            tag = XmpTag(key, tag_or_value)

        tag._set_owner(self)
        self._tags['xmp'][tag.key] = tag
        if tag.key not in self.xmp_keys:
            self._keys['xmp'].append(tag.key)

    def __setitem__(self, key, tag_or_value):
        """Set a metadata tag for a given key.

        If the tag was previously set, it is overwritten.
        As a handy shortcut, a value may be passed instead of a fully formed
        tag. The corresponding tag object will be instantiated.

        Raise KeyError if the key is invalid

        Args:
        key -- metadata key in the dotted form
               ``familyName.groupName.tagName`` where ``familyName`` may
               be one of ``exif``, ``iptc`` or ``xmp``.
        tag_or_value -- an instance of the corresponding family of metadata
                        tag or a value
                        Type: pyexiv2.exif.ExifTag instance or
                              pyexiv2.iptc.IptcTag instance or
                              pyexiv2.xmp.XmpTag instance or 
                              any valid value type
        """
        family = key.split('.')[0].lower()
        if family in ('exif', 'iptc', 'xmp'):
            return getattr(self, '_set_%s_tag' % family)(key, tag_or_value)

        else:
            raise KeyError(key)

    def _delete_exif_tag(self, key):
        """Delete an EXIF tag.

        Throw a KeyError if the tag doesn't exist.

        Args:
        key -- the EXIF key
        """
        if key not in self.exif_keys:
            raise KeyError('Cannot delete an inexistent tag')

        self._image._deleteExifTag(key)
        try:
            del self._tags['exif'][key]
        except KeyError:
            # The tag was not cached.
            pass

        if self._keys['exif'] is not None:
            self._keys['exif'].remove(key)

    def _delete_iptc_tag(self, key):
        """Delete an IPTC tag.

        Throw a KeyError if the tag doesn't exist.

        Args:
        key -- the IPTC key
        """
        if key not in self.iptc_keys:
            raise KeyError('Cannot delete an inexistent tag')

        self._image._deleteIptcTag(key)
        try:
            del self._tags['iptc'][key]
        except KeyError:
            # The tag was not cached.
            pass

        if self._keys['iptc'] is not None:
            self._keys['iptc'].remove(key)

    def _delete_xmp_tag(self, key):
        """Delete an XMP tag.

        Throw a KeyError if the tag doesn't exist.

        Args:
        key -- the XMP key
        """
        if key not in self.xmp_keys:
            raise KeyError('Cannot delete an inexistent tag')

        self._image._deleteXmpTag(key)
        try:
            del self._tags['xmp'][key]
        except KeyError:
            # The tag was not cached.
            pass

        if self._keys['xmp'] is not None:
            self._keys['xmp'].remove(key)

    def __delitem__(self, key):
        """Delete a metadata tag for a given key.

        Raise KeyError if the tag with the given key doesn't exist

        Args:
        key -- the metadata key in the dotted form
               ``familyName.groupName.tagName`` where ``familyName`` may
               be one of ``exif``, ``iptc`` or ``xmp``.
        """
        family = key.split('.')[0].lower()
        if family in ('exif', 'iptc', 'xmp'):
            return getattr(self, '_delete_%s_tag' % family)(key)

        else:
            raise KeyError(key)

    def __iter__(self):
        return chain(self.exif_keys, self.iptc_keys, self.xmp_keys)

    def __len__(self):
        return len([x for x in self])

    def _get_comment(self):
        return self._image._getComment()

    def _set_comment(self, comment):
        if comment is not None:
            self._image._setComment(comment)

        else:
            self._del_comment()

    def _del_comment(self):
        self._image._clearComment()

    comment = property(fget=_get_comment, fset=_set_comment, fdel=_del_comment,
                       doc='The image comment.')

    @property
    def previews(self):
        """List of the previews available in the image, sorted by increasing
        size.

        """
        return [Preview(preview) for preview in self._image._previews()]

    def copy(self, other, exif=True, iptc=True, xmp=True, comment=True):
        """Copy the metadata to another image.

        The metadata in the destination is overridden. In particular, if the
        destination contains e.g. EXIF data and the source doesn't, it will be
        erased in the destination, unless explicitly omitted.

        Args:
        other -- the destination metadata to copy to (it must have been
                 read beforehand)
                 Type: pyexiv2.metadata.ImageMetadata instance
        exif -- whether to copy the EXIF metadata, default True
        iptc -- whether to copy the IPTC metadata, default True
        xmp -- whether to copy the XMP metadata, default True
        comment -- whether to copy the image comment, default True 
        """
        self._image._copyMetadata(other._image, exif, iptc, xmp)
        # Empty the cache where needed
        if exif:
            other._keys['exif'] = None
            other._tags['exif'] = {}

        if iptc:
            other._keys['iptc'] = None
            other._tags['iptc'] = {}

        if xmp:
            other._keys['xmp'] = None
            other._tags['xmp'] = {}

        if comment:
            other.comment = self.comment

    @property
    def buffer(self):
        """
        The image buffer as a string.
        If metadata has been modified, the data won't be up-to-date until
        :meth:`.write` has been called.
        """
        return self._image._getDataBuffer()

    @property
    def exif_thumbnail(self):
        """A thumbnail image optionally embedded in the EXIF data.

        """
        if self._exif_thumbnail is None:
            self._exif_thumbnail = ExifThumbnail(self)

        return self._exif_thumbnail

    def _get_iptc_charset(self):
        value = self._image._getIptcCharset()
        if value != '':
            return value.lower()

        else:
            return None

    def _set_iptc_charset(self, charset):
        if charset is None:
            self._del_iptc_charset()
            return

        try:
            name = codecs.lookup(charset).name
        except LookupError as error:
            raise ValueError(error)

        else:
            charsets = {'utf-8': '\x1b%G'}
            try:
                self['Iptc.Envelope.CharacterSet'] = (charsets[name],)
            except KeyError:
                raise ValueError('Unhandled charset: %s' % name)

    def _del_iptc_charset(self):
        try:
            del self['Iptc.Envelope.CharacterSet']
        except KeyError:
            pass

    iptc_charset = property(fget=_get_iptc_charset, 
                            fset=_set_iptc_charset,
                            fdel=_del_iptc_charset,
                            doc='An optional character set the IPTC data'\
                                ' is encoded in.')

    # Some convenient functions -------------------------------------------
    def get_iso(self):
        """Returns the ISO value as integer.

        """
        try:
            return self["Exif.Photo.ISOSpeedRatings"].value
        except KeyError:
            return

    def get_shutter_speed(self, float_=False):
        """Returns the exposure time as rational or float.

        Args:
        float_ -- if False, default, the value is returned as rational 
                  otherwise as float
        """
        try:
            speed = self['Exif.Photo.ExposureTime'].value 
        except KeyError:
            return

        if float_:
            if speed.denominator:
                return speed.numerator / speed.denominator
            return float(numerator)

        return speed

    def get_focal_length(self):
        """Returns the focal length as float.

        """
        try:
            focal = self['Exif.Photo.FocalLength'].value
            if focal.denominator:
                return round(focal.numerator / focal.denominator, 2)

            else:
                return float(focal.numerator)

        except KeyError:
            return

    def get_aperture(self):
        """Returns the fNumber as float.

        """
        try:
            fnumber = self["Exif.Photo.FNumber"].value
            if fnumber.denominator:
                return round(fnumber.numerator / fnumber.denominator, 2)

            else:
                return float(fnumber.numerator)
        except KeyError:
            return

    def get_orientation(self):
        """Returns the orientation of the image as integer.

        If the tag is not set, the value 1 is returned.
        """
        try:
            return self["Exif.Image.Orientation"].value
        except KeyError:
            return 1

    def get_exposure_data(self, float_=False):
        """Returns the exposure parameters of the image.

        The values are returned as a dict which contains:
            "iso": the ISO value
            "speed": the exposure time
            "focal": the focal length
            "aperture": the fNumber
            "orientation": the orientation of the image

        When a tag is not set, the value will be None

        Args:
        float_ -- if False, default, the value of the exposure time is returned 
                  as rational otherwise as float
        """
        data = {"iso": self.get_iso(),
                "speed": self.get_shutter_speed(float_),
                "focal": self.get_focal_length(),
                "aperture": self.get_aperture(),
                "orientation": self.get_orientation()}

        return data

    def get_rights_data(self):
        """Returns the author and copyright info.

        The values are returned as a dict which contains:
            "creator": the value of Xmp.dc.creator
            "artist": the value of Exif.Image.Artist
            "rights": the value of Xmp.dc.rights
            "copyright": the value of Exif.Image.Copyright
            "marked": the value of Xmp.xmpRights.Marked
            "usage": the value of Xmp.xmpRights.UsageTerms

        When a tag is not set, the value will be None
        """
        tags = [('creator', 'Xmp.dc.creator'), 
                ('artist', 'Exif.Image.Artist'),
                ('rights', 'Xmp.dc.rights'), 
                ('copyright', 'Exif.Image.Copyright'),
                ('marked', 'Xmp.xmpRights.Marked'), 
                ('usage', 'Xmp.xmpRights.UsageTerms')]
        rights = {}
        for tag in tags:
            try:
                rights[tag[0]] = self[tag[1]].value
            except KeyError:
                rights[tag[0]] = None

        return rights





# -*- coding: utf-8 -*-

# ******************************************************************************
#
# Copyright (C) 2006-2012 Olivier Tilloy <olivier@tilloy.net>
# Copyright (C) 2015-2023 Vincent Vande Vyvre <vincent.vandevyvre@oqapy.eu>
#
# This file is part of the py3exiv2 distribution.
#
# py3exiv2 is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 3 as published by the Free Software Foundation.
#
# py3exiv2 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with py3exiv2; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, 5th Floor, Boston, MA 02110-1301 USA.
#
# Maintainer: Vincent Vande Vyvre <vincent.vandevyvre@oqapy.eu>
#
# ******************************************************************************

"""
Utilitary classes and functions.
"""

import datetime
import re

from fractions import Fraction

class FixedOffset(datetime.tzinfo):
    """Define a fixed positive or negative offset of a local time from UTC.

    """

    def __init__(self, sign='+', hours=0, minutes=0):
        """Initialize an offset from a sign ('+' or '-') and an absolute value
        expressed in hours and minutes.

        No check on the validity of those values is performed, it is the
        responsibility of the caller to pass valid values.

        Args:
        sign -- the sign of the offset ('+' or '-')
        hours -- an absolute number of hours
        minutes -- an absolute number of minutes
        """
        self.sign = sign
        self.hours = hours
        self.minutes = minutes

    def utcoffset(self, dt):
        """Return offset of local time from UTC, in minutes east of UTC.

        If local time is west of UTC, this value will be negative.

        Args:
        dt -- the local datetime.time instance

        Return: a whole number of minutes in the range -1439 to 1439 inclusive
        """
        total = self.hours * 60 + self.minutes
        if self.sign == '-':
            total = -total
        return datetime.timedelta(minutes = total)

    def dst(self, dt):
        """Return the daylight saving time (DST) adjustment.

        In this implementation, it is always nil.

        Args:
        dt -- the local datetime.time instance

        Return: the DST adjustment (always nil)
        """
        return datetime.timedelta(0)

    def tzname(self, dt):
        """Return a string representation of the offset in the format '±%H:%M'.

        If the offset is nil, the representation is, by convention, 'Z'.

        Args:
        dt -- the local datetime.time instance

        Return: a human-readable representation of the offset
        """
        if self.hours == 0 and self.minutes == 0:
            return 'Z'
        else:
            return '%s%02d:%02d' % (self.sign, self.hours, self.minutes)

    def __equal__(self, other):
        """Test equality between this offset and another offset.

        Args:
        other -- another FixedOffset instance

        Return: True if the offset are equal, False otherwise
        """
        return (self.sign == other.sign) and (self.hours == other.hours) and \
            (self.minutes == other.minutes)


def undefined_to_string(undefined):
    """Convert an undefined string into its corresponding sequence of bytes.

    The undefined string must contain the ascii codes of a sequence of bytes,
    separated by white spaces (e.g. "48 50 50 49" will be converted into
    "0221").
    The Undefined type is part of the EXIF specification.

    Args:
    undefined -- an undefined string

    Return: the corresponding decoded string
    """
    if not undefined:
        return ''

    return ''.join([chr(int(x)) for x in undefined.rstrip().split(' ')])


def string_to_undefined(sequence):
    """Convert a string into its undefined form.

    The undefined form contains a sequence of ascii codes separated by white
    spaces (e.g. "0221" will be converted into "48 50 50 49").
    The Undefined type is part of the EXIF specification.

    Args:
    sequence -- a sequence of bytes

    Return: the corresponding undefined string
    """
    return " ".join([str(ord(s)) for s in sequence])

def is_fraction(obj):
    """Test whether the object is a valid fraction.

    """
    return isinstance(obj, Fraction)

def match_string(string):
    """Match a string against the expected format for a :class:`Fraction`
    (``[-]numerator/denominator``) and return the numerator and denominator
    as a tuple.

    Args: 
    string -- a string representation of a rational number

    Return: a tuple (numerator, denominator)

    Raise ValueError: if the format of the string is invalid
    """
    format_re = re.compile(r'(?P<numerator>-?\d+)/(?P<denominator>\d+)')
    match = format_re.match(string)
    if match is None:
        raise ValueError('Invalid format for a rational: %s' % string)

    gd = match.groupdict()
    return (int(gd['numerator']), int(gd['denominator']))

def make_fraction(*args):
    """Make a fraction.

    Raise TypeError: if the arguments do not match the expected format for a
                      fraction
    """
    if len(args) == 1:
        numerator, denominator = match_string(args[0])

    elif len(args) == 2:
        numerator = args[0]
        denominator = args[1]

    else:
        raise TypeError('Invalid format for a fraction: %s' % str(args))

    if denominator == 0 and numerator == 0:
        # Null rationals are often stored as '0/0'.
        # We want to be fault-tolerant in this specific case
        # (see https://bugs.launchpad.net/pyexiv2/+bug/786253).
        denominator = 1

    return Fraction(numerator, denominator)


def fraction_to_string(fraction):
    """Return a string representation of a fraction.

    The returned string is always in the form '[numerator]/[denominator]'.

    Raise TypeError: if the argument is not a valid fraction
    """
    if isinstance(fraction, Fraction):
        # fractions.Fraction.__str__ returns '0' for a null numerator.
        return '%s/%s' % (fraction.numerator, fraction.denominator)

    else:
        raise TypeError('Not a fraction')


class ListenerInterface(object):

    """
    Interface that an object that wants to listen to changes on another object
    should implement.
    """

    def contents_changed(self):
        """
        React on changes on the object observed.
        Override to implement specific behaviours.
        """
        raise NotImplementedError()


class NotifyingList(list):
    """A simplistic implementation of a notifying list.

    Any changes to the list are notified in a synchronous way to all previously
    registered listeners. 
    A listener must implement the class ListenerInterface.
    """

    # Useful documentation:
    # file:///usr/share/doc/python2.5/html/lib/typesseq-mutable.html
    # http://docs.python.org/reference/datamodel.html#additional-methods-for-emulation-of-sequence-types

    def __init__(self, items=[]):
        super(NotifyingList, self).__init__(items)
        self._listeners = set()

    def register_listener(self, listener):
        """Register a new listener to be notified of changes.

        Args:
        listener -- any ListenerInterface instance that listens for changes
        """
        self._listeners.add(listener)

    def unregister_listener(self, listener):
        """Unregister a previously registered listener.

        Args:
        listener -- a previously registered listener

        Raise KeyError: if the listener was not previously registered
        """
        self._listeners.remove(listener)

    def _notify_listeners(self, *args):
        for listener in self._listeners:
            listener.contents_changed(*args)

    def __setitem__(self, index, item):
        # FIXME: support slice arguments for extended slicing
        super(NotifyingList, self).__setitem__(index, item)
        self._notify_listeners()

    def __delitem__(self, index):
        # FIXME: support slice arguments for extended slicing
        super(NotifyingList, self).__delitem__(index)
        self._notify_listeners()

    def append(self, item):
        super(NotifyingList, self).append(item)
        self._notify_listeners()

    def extend(self, items):
        super(NotifyingList, self).extend(items)
        self._notify_listeners()

    def insert(self, index, item):
        super(NotifyingList, self).insert(index, item)
        self._notify_listeners()

    def pop(self, index=None):
        if index is None:
            item = super(NotifyingList, self).pop()
        else:
            item = super(NotifyingList, self).pop(index)
        self._notify_listeners()
        return item

    def remove(self, item):
        super(NotifyingList, self).remove(item)
        self._notify_listeners()

    def reverse(self):
        super(NotifyingList, self).reverse()
        self._notify_listeners()

    def sort(self, key=None, reverse=False):
        super(NotifyingList, self).sort(key=key, reverse=reverse)
        self._notify_listeners()

    def __iadd__(self, other):
        self = super(NotifyingList, self).__iadd__(other)
        self._notify_listeners()
        return self

    def __imul__(self, coefficient):
        self = super(NotifyingList, self).__imul__(coefficient)
        self._notify_listeners()
        return self

    def setslice__(self, i, j, items):
        # __setslice__ is deprecated but needs to be overridden for completeness
        super(NotifyingList, self).__setslice__(i, j, items)
        self._notify_listeners()

    def delslice__(self, i, j):
        # __delslice__ is deprecated but needs to be overridden for completeness
        deleted = self[i:j]
        super(NotifyingList, self).__delslice__(i, j)
        if deleted:
            self._notify_listeners()


class GPSCoordinate(object):
    """A class representing GPS coordinates (e.g. a latitude or a longitude).

    Its attributes (degrees, minutes, seconds, direction) are read-only
    properties.
    """
    _format_re = re.compile(r'(?P<degrees>-?\d+),'
                    '(?P<minutes>\d+)(,(?P<seconds>\d+)|\.(?P<fraction>\d+))'
                    '(?P<direction>[NSEW])')

    def __init__(self, degrees, minutes, seconds, direction):
        """Instanciate a GPSCoordinate object.

        Args:
        degrees -- int(degrees)
        minutes -- int(minutes)
        seconds -- int(seconds)
        direction -- str('N', 'S', 'E' or 'W')

        Raise ValueError: if any of the parameter is not in the expected range
                           of values
        """
        if direction not in ('N', 'S', 'E', 'W'):
            raise ValueError('Invalid direction: %s' % direction)

        self._direction = direction
        if (direction in ('N', 'S') and (degrees < 0 or degrees > 90)) or \
           (direction in ('E', 'W') and (degrees < 0 or degrees > 180)):
            raise ValueError('Invalid value for degrees: %d' % degrees)

        self._degrees = degrees
        if minutes < 0 or minutes > 60:
            raise ValueError('Invalid value for minutes: %d' % minutes)

        self._minutes = minutes
        if seconds < 0 or seconds > 60:
            raise ValueError('Invalid value for seconds: %d' % seconds)

        self._seconds = seconds

    @property
    def degrees(self):
        """The degrees component of the coordinate.

        """
        return self._degrees

    @property
    def minutes(self):
        """The minutes component of the coordinate.

        """
        return self._minutes

    @property
    def seconds(self):
        """The seconds component of the coordinate.

        """
        return self._seconds

    @property
    def direction(self):
        """The direction component of the coordinate.

        """
        return self._direction

    @staticmethod
    def from_string(string):
        """Instantiate a :class:`GPSCoordinate` from a string formatted as
        ``DDD,MM,SSk`` or ``DDD,MM.mmk`` where ``DDD`` is a number of degrees,
        ``MM`` is a number of minutes, ``SS`` is a number of seconds, ``mm`` is
        a fraction of minutes, and ``k`` is a single character N, S, E, W
        indicating a direction (north, south, east, west).

        Args:
        string -- a string representation of a GPS coordinate

        Return: the GPSCoordinate parsed

        Raise ValueError: if the format of the string is invalid
        """
        match = GPSCoordinate._format_re.match(string)
        if match is None:
            raise ValueError('Invalid format for a GPS coordinate: %s' % string)

        gd = match.groupdict()
        fraction = gd['fraction']
        if fraction is not None:
            seconds = int(round(int(fraction[:2]) * 0.6))

        else:
            seconds = int(gd['seconds'])

        return GPSCoordinate(int(gd['degrees']), int(gd['minutes']), seconds,
                             gd['direction'])

    def __eq__(self, other):
        """Compare two GPS coordinates for equality.

        Two coordinates are equal if and only if all their components are equal.

        Args:
        other -- the GPSCoordinate instance to compare to self for equality

        Return: True if equal, False otherwise
        """
        return (self._degrees == other._degrees) and \
               (self._minutes == other._minutes) and \
               (self._seconds == other._seconds) and \
               (self._direction == other._direction)

    def __str__(self):
        """Return a string representation of the GPS coordinate conforming
        to the XMP specification

        """
        return '%d,%d,%d%s' % (self._degrees, self._minutes, self._seconds,
                               self._direction)


class DateTimeFormatter(object):
    """Convenience object that exposes static methods to convert a date, time or
    datetime object to a string representation suitable for various metadata
    standards.

    This is needed because python’s
    `strftime() <http://docs.python.org/library/datetime.html#strftime-strptime-behavior>`_
    doesn’t work for years before 1900.

    This class mostly exists for internal usage only. Clients should never need
    to use it.
    """

    @staticmethod
    def timedelta_to_offset(t):
        """Convert a time delta to a string..

        Args:
        t -- a datetime.timedelta instance

        Return: a string representation of the time delta in the form `±%H:%M`
        """
        # timedelta.total_seconds() is only available starting with Python 3.2
        seconds = t.total_seconds()
        hours = int(seconds / 3600)
        minutes = abs(int((seconds - hours * 3600) / 60))
        return '%+03d:%02d' % (hours, minutes)

    @staticmethod
    def exif(d):
        """Convert a date/time object to a string representation conforming to
        libexiv2’s internal representation for the EXIF standard.

        Args:
        d -- a datetime.datetime or datetime.date instance

        Return: a string representation conforming to the EXIF standard

        Raise TypeError: if the parameter is not a datetime or a date object
        """
        if isinstance(d, datetime.datetime):
            return '%04d:%02d:%02d %02d:%02d:%02d' % \
                (d.year, d.month, d.day, d.hour, d.minute, d.second)

        elif isinstance(d, datetime.date):
            return '%04d:%02d:%02d' % (d.year, d.month, d.day)

        else:
            raise TypeError('expecting an object of type '
                            'datetime.datetime or datetime.date')

    @staticmethod
    def iptc_date(d):
        """Convert a date object to a string representation conforming to
        libexiv2’s internal representation for the IPTC standard.

        Args:
        d -- a datetime.datetime or datetime.date instance

        Return: a string representation conforming to the IPTC standard

        Raise TypeError: if the parameter is not a datetime or a date object
        """
        if isinstance(d, (datetime.date, datetime.datetime)):
            # ISO 8601 date format.
            # According to the IPTC specification, the format for a string
            # field representing a date is '%Y%m%d'. However, the string
            # expected by exiv2's DateValue::read(string) should be
            # formatted using pattern '%Y-%m-%d'.
            return '%04d-%02d-%02d' % (d.year, d.month, d.day)

        else:
            raise TypeError('expecting an object of type '
                            'datetime.datetime or datetime.date')

    @staticmethod
    def iptc_time(d):
        """Convert a time object to a string representation conforming to
        libexiv2’s internal representation for the IPTC standard.

        Args:
        d -- a datetime.datetime or time instance

        Return: a string representation conforming to the IPTC standard

        Raise TypeError: if the parameter is not a datetime or a time object
        """
        if isinstance(d, (datetime.time, datetime.datetime)):
            # According to the IPTC specification, the format for a string
            # field representing a time is '%H%M%S±%H%M'. However, the
            # string expected by exiv2's TimeValue::read(string) should be
            # formatted using pattern '%H:%M:%S±%H:%M'.
            r = '%02d:%02d:%02d' % (d.hour, d.minute, d.second)
            if d.tzinfo is not None:
                t = d.utcoffset()
                if t is not None:
                    r += DateTimeFormatter.timedelta_to_offset(t)

            else:
                r += '+00:00'

            return r

        else:
            raise TypeError('expecting an object of type '
                            'datetime.datetime or datetime.time')

    @staticmethod
    def xmp(d):
        """Convert a date/time object to a string representation conforming to
        libexiv2’s internal representation for the XMP standard.

        Args:
        d -- a datetime.datetime or datetime.date instance

        Return: a string representation conforming to the XMP standard

        Raise TypeError: if the parameter is not a datetime or a date object
        """
        if isinstance(d, datetime.datetime):
            t = d.utcoffset()
            if d.tzinfo is None or t is None or t == datetime.timedelta(0):
                tz = 'Z'

            else:
                tz = DateTimeFormatter.timedelta_to_offset(t)

            if d.hour == 0 and d.minute == 0 and \
                d.second == 0 and d.microsecond == 0 and \
                (d.tzinfo is None or d.utcoffset() == datetime.timedelta(0)):
                return '%04d-%02d-%02d' % (d.year, d.month, d.day)

            elif d.second == 0 and d.microsecond == 0:
                return '%04d-%02d-%02dT%02d:%02d%s' % \
                    (d.year, d.month, d.day, d.hour, d.minute, tz)

            elif d.microsecond == 0:
                return '%04d-%02d-%02dT%02d:%02d:%02d%s' % \
                    (d.year, d.month, d.day, d.hour, d.minute, d.second, tz)

            else:
                r = '%04d-%02d-%02dT%02d:%02d:%02d.' % \
                    (d.year, d.month, d.day, d.hour, d.minute, d.second)
                r += str(int(d.microsecond) / 1E6)[2:]
                r += tz
                return r

        elif isinstance(d, datetime.date):
            return '%04d-%02d-%02d' % (d.year, d.month, d.day)

        else:
            raise TypeError('expecting an object of type '
                            'datetime.datetime or datetime.date')

